import argparse
import os
from typing import List

from ..ui.utils import create_lab_table
from ..ui.utils import create_panel, LabMetaHighlighter
from ... import utils
from ...exceptions import PrivilegeError, EmptyLabError
from ...foundation.cli.command.Command import Command
from ...manager.Kathara import Kathara
from ...parser.netkit.DepParser import DepParser
from ...parser.netkit.ExtParser import ExtParser
from ...parser.netkit.FolderParser import FolderParser
from ...parser.netkit.LabParser import LabParser
from ...parser.netkit.OptionParser import OptionParser
from ...setting.Setting import Setting
from ...strings import strings, wiki_description


class LstartCommand(Command):
    def __init__(self) -> None:
        Command.__init__(self)

        self.parser: argparse.ArgumentParser = argparse.ArgumentParser(
            prog='kathara lstart',
            description=strings['lstart'],
            epilog=wiki_description,
            add_help=False
        )

        self.parser.add_argument(
            '-h', '--help',
            action='help',
            default=argparse.SUPPRESS,
            help='Show a help message and exit.'
        )

        group = self.parser.add_mutually_exclusive_group(required=False)

        group.add_argument(
            "--noterminals",
            action="store_const",
            dest="terminals",
            const=False,
            default=None,
            help='Start the network scenario without opening terminal windows.'
        )
        group.add_argument(
            "--terminals",
            action="store_const",
            dest="terminals",
            const=True,
            help='Start the network scenario opening terminal windows.'
        )
        self.parser.add_argument(
            "--privileged",
            action="store_const",
            const=True,
            required=False,
            help='Start the devices in privileged mode. MUST BE ROOT FOR THIS OPTION.'
        )
        self.parser.add_argument(
            '-d', '--directory',
            required=False,
            help='Specify the folder containing the network scenario.'
        )
        self.parser.add_argument(
            '-F', '--force-lab',
            dest='force_lab',
            required=False,
            action='store_true',
            help='Force the network scenario to start without a lab.conf or lab.dep file.'
        )
        self.parser.add_argument(
            '-l', '--list',
            required=False,
            action='store_true',
            help='Show information about running devices after the network scenario has been started.'
        )
        self.parser.add_argument(
            '-o', '--pass',
            dest='global_machine_metadata',
            metavar="METADATA",
            nargs='*',
            required=False,
            help="Apply metadata to all devices of a network scenario during startup."
        )
        self.parser.add_argument(
            '--terminal-emu',
            required=False,
            help='Set a different terminal emulator application (Unix only).'
        )
        self.parser.add_argument(
            '--print', '--dry-mode',
            dest="dry_mode",
            required=False,
            action='store_true',
            help='Open the lab.conf file and check if it is correct (dry run).'
        )
        hosthome_group = self.parser.add_mutually_exclusive_group(required=False)
        hosthome_group.add_argument(
            '--no-hosthome', '-H',
            dest="hosthome_mount",
            action="store_const",
            const=False,
            help='Do not mount "/hosthome" directory inside devices.'
        )
        hosthome_group.add_argument(
            '--hosthome',
            dest="hosthome_mount",
            action="store_const",
            const=True,
            help='Mount "/hosthome" directory inside devices.'
        )
        shared_group = self.parser.add_mutually_exclusive_group(required=False)
        shared_group.add_argument(
            '--no-shared', '-S',
            dest="shared_mount",
            action="store_const",
            const=False,
            help='Do not mount "/shared" directory inside devices.'
        )
        shared_group.add_argument(
            '--shared',
            dest="shared_mount",
            action="store_const",
            const=True,
            help='Mount "/shared" directory inside devices.'
        )
        self.parser.add_argument(
            '--exclude',
            dest='excluded_machines',
            metavar='DEVICE_NAME',
            nargs='+',
            default=[],
            help='Exclude specified devices from startup.'
        )
        self.parser.add_argument(
            'machine_name',
            metavar='DEVICE_NAME',
            nargs='*',
            help='Launches only specified devices.'
        )

    def run(self, current_path: str, argv: List[str]) -> int:
        self.parse_args(argv)
        args = self.get_args()

        lab_path = args['directory'].replace('"', '').replace("'", '') if args['directory'] else current_path
        lab_path = utils.get_absolute_path(lab_path)

        # Load custom 'kathara.conf' if it exists
        self._load_custom_configuration(lab_path)

        Setting.get_instance().open_terminals = args['terminals'] if args['terminals'] is not None \
            else Setting.get_instance().open_terminals
        Setting.get_instance().terminal = args['terminal_emu'] or Setting.get_instance().terminal

        self.console.print(
            create_panel(
                "Checking Network Scenario" if args['dry_mode'] else "Starting Network Scenario",
                style="blue bold", justify="center"
            )
        )

        try:
            lab = LabParser.parse(lab_path)
        except IOError as e:
            if not args['force_lab']:
                raise e
            else:
                lab = FolderParser.parse(lab_path)

        # Reorder machines by lab.dep file, if present.
        dependencies = DepParser.parse(lab_path)
        if dependencies:
            lab.apply_dependencies(dependencies)

        lab_meta_information = str(lab)
        if lab_meta_information:
            meta_highlighter = LabMetaHighlighter()
            self.console.print(create_panel(meta_highlighter(lab_meta_information)))

        if len(lab.machines) <= 0:
            raise EmptyLabError()

        lab.global_machine_metadata = OptionParser.parse(args['global_machine_metadata'])

        lab_ext_path = os.path.join(lab_path, 'lab.ext')
        lab_ext_exists = False
        if os.path.exists(lab_ext_path):
            lab_ext_exists = True
            if utils.is_platform(utils.LINUX) or utils.is_platform(utils.LINUX2):
                if utils.is_admin():
                    external_links = ExtParser.parse(lab_path)

                    if external_links:
                        lab.attach_external_links(external_links)

                        # Since xterm does not work with "sudo", we do not open terminals when lab.ext is present.
                        Setting.get_instance().open_terminals = False
                else:
                    raise PrivilegeError("You must be root in order to use lab.ext file.")
            else:
                raise OSError("lab.ext is only available on Linux systems.")

        # If dry mode, we just check if the lab.conf is correct.
        if args['dry_mode']:
            self.console.print("[green]\u2713 [bold]lab.conf[/bold] file is correct.")
            if dependencies:
                self.console.print("[green]\u2713 [bold]lab.dep[/bold] file is correct.")
            if lab_ext_exists:
                self.console.print("[green]\u2713 [bold]lab.ext[/bold] file is correct.")

            return 0

        lab.add_option('hosthome_mount', args['hosthome_mount'])
        lab.add_option('shared_mount', args['shared_mount'])

        if args['privileged'] or any(x.is_privileged() for x in lab.machines.values()):
            if not utils.is_admin():
                raise PrivilegeError("You must be root in order to start Kathara devices in privileged mode.")
            else:
                if Setting.get_instance().open_terminals:
                    self.console.print(
                        "[yellow]\u26a0 Running devices with privileged capabilities, terminals might not open!"
                    )
        lab.add_global_machine_metadata('privileged', args['privileged'])

        Kathara.get_instance().deploy_lab(
            lab, selected_machines=set(args['machine_name']), excluded_machines=set(args['excluded_machines'])
        )

        if args['list']:
            with self.console.status(
                    f"Loading...",
                    spinner="dots"
            ) as _:
                machines_stats = Kathara.get_instance().get_machines_stats(lab_hash=lab.hash)
                self.console.print(create_lab_table(machines_stats))

        return 0



import io
from abc import ABC, abstractmethod
from typing import Dict, Set, Any, Generator, Tuple, List, Optional, Union

from .exec_stream.IExecStream import IExecStream
from .stats.ILinkStats import ILinkStats
from .stats.IMachineStats import IMachineStats
from ...model.Lab import Lab
from ...model.Link import Link
from ...model.Machine import Machine


class IManager(ABC):
    """Interface to be implemented in the virtualization managers"""

    @abstractmethod
    def deploy_machine(self, machine: Machine) -> None:
        """Deploy a Kathara device.

        Args:
            machine (Kathara.model.Machine): A Kathara machine object.

        Returns:
            None

        Raises:
            LabNotFoundError: If the specified device is not associated to any network scenario.
        """
        raise NotImplementedError("You must implement `deploy_machine` method.")

    @abstractmethod
    def deploy_link(self, link: Link) -> None:
        """Deploy a Kathara collision domain.

        Args:
            link (Kathara.model.Link): A Kathara collision domain object.

        Returns:
            None

        Raises:
            LabNotFoundError: If the collision domain is not associated to any network scenario.
        """
        raise NotImplementedError("You must implement `deploy_link` method.")

    @abstractmethod
    def deploy_lab(self, lab: Lab, selected_machines: Optional[Set[str]] = None,
                   excluded_machines: Optional[Set[str]] = None) -> None:
        """Deploy a Kathara network scenario.

        Args:
            lab (Kathara.model.Lab): A Kathara network scenario.
            selected_machines (Optional[Set[str]]): If not None, deploy only the specified devices.
            excluded_machines (Optional[Set[str]]): If not None, exclude devices from being deployed.

        Returns:
            None
        """
        raise NotImplementedError("You must implement `deploy_lab` method.")

    @abstractmethod
    def connect_machine_to_link(self, machine: Machine, link: Link, mac_address: Optional[str] = None) -> None:
        """Connect a Kathara device to a collision domain.

        Args:
            machine (Kathara.model.Machine): A Kathara machine object.
            link (Kathara.model.Link): A Kathara collision domain object.
            mac_address (Optional[str]): The MAC address to assign to the interface.

        Returns:
            None

        Raises:
            LabNotFoundError: If the device specified is not associated to any network scenario.
            LabNotFoundError: If the collision domain is not associated to any network scenario.
            MachineCollisionDomainConflictError: If the device is already connected to the collision domain.
        """
        raise NotImplementedError("You must implement `connect_machine_to_link` method.")

    def disconnect_machine_from_link(self, machine: Machine, link: Link) -> None:
        """Disconnect a Kathara device from a collision domain.

        Args:
            machine (Kathara.model.Machine): A Kathara machine object.
            link (Kathara.model.Link): The Kathara collision domain from which disconnect the device.

        Returns:
            None

        Raises:
            LabNotFoundError: If the device specified is not associated to any network scenario.
            LabNotFoundError: If the collision domain is not associated to any network scenario.
            MachineCollisionDomainConflictError: If the device is not connected to the collision domain.
        """
        raise NotImplementedError("You must implement `disconnect_machine_from_link` method.")

    @abstractmethod
    def undeploy_machine(self, machine: Machine) -> None:
        """Undeploy a Kathara device.

        Args:
            machine (Kathara.model.Machine): A Kathara machine object.

        Returns:
            None

        Raises:
            LabNotFoundError: If the device specified is not associated to any network scenario.
        """
        raise NotImplementedError("You must implement `undeploy_machine` method.")

    @abstractmethod
    def undeploy_link(self, link: Link) -> None:
        """Undeploy a Kathara collision domain.

        Args:
            link (Kathara.model.Link): A Kathara collision domain object.

        Returns:
            None

        Raises:
            LabNotFoundError: If the collision domain is not associated to any network scenario.
        """
        raise NotImplementedError("You must implement `undeploy_link` method.")

    @abstractmethod
    def undeploy_lab(self, lab_hash: Optional[str] = None, lab_name: Optional[str] = None, lab: Optional[Lab] = None,
                     selected_machines: Optional[Set[str]] = None,
                     excluded_machines: Optional[Set[str]] = None) -> None:
        """Undeploy a Kathara network scenario.

        Args:
            lab_hash (Optional[str]): The hash of the network scenario.
                Can be used as an alternative to lab_name and lab. If None, lab_name or lab should be set.
            lab_name (Optional[str]): The name of the network scenario.
                Can be used as an alternative to lab_hash and lab. If None, lab_hash or lab should be set.
            lab (Optional[Kathara.model.Lab]): The network scenario object.
                Can be used as an alternative to lab_hash and lab_name. If None, lab_hash or lab_name should be set.
            selected_machines (Optional[Set[str]]): If not None, undeploy only the specified devices.
            excluded_machines (Optional[Set[str]]): If not None, exclude devices from being undeployed.

        Returns:
            None

        Raises:
            InvocationError: If a running network scenario hash or name is not specified.
        """
        raise NotImplementedError("You must implement `undeploy_lab` method.")

    @abstractmethod
    def wipe(self, all_users: bool = False) -> None:
        """Undeploy all the running network scenarios.

        Args:
            all_users (bool): If false, undeploy only the current user network scenarios. If true, undeploy the
                running network scenarios of all users.

        Returns:
            None
        """
        raise NotImplementedError("You must implement `wipe` method.")

    @abstractmethod
    def connect_tty(self, machine_name: str, lab_hash: Optional[str] = None, lab_name: Optional[str] = None,
                    lab: Optional[Lab] = None, shell: str = None, logs: bool = False,
                    wait: Union[bool, Tuple[int, float]] = True) -> None:
        """Connect to a device in a running network scenario, using the specified shell.

        Args:
            machine_name (str): The name of the device to connect.
            lab_hash (Optional[str]): The hash of the network scenario.
                Can be used as an alternative to lab_name and lab. If None, lab_name or lab should be set.
            lab_name (Optional[str]): The name of the network scenario.
                Can be used as an alternative to lab_hash and lab. If None, lab_hash or lab should be set.
            lab (Optional[Kathara.model.Lab]): The network scenario object.
                Can be used as an alternative to lab_hash and lab_name. If None, lab_hash or lab_name should be set.
            shell (str): The name of the shell to use for connecting.
            logs (bool): If True, print startup logs on stdout.
            wait (Union[bool, Tuple[int, float]]): If True, wait indefinitely until the end of the startup commands
                execution before connecting. If a tuple is provided, the first value indicates the number of retries
                before stopping waiting and the second value indicates the time interval to wait for each retry.
                Default is True.

        Returns:
            None

        Raises:
            InvocationError: If a running network scenario hash or name is not specified.
        """
        raise NotImplementedError("You must implement `connect_tty` method.")

    @abstractmethod
    def connect_tty_obj(self, machine: Machine, shell: str = None, logs: bool = False,
                        wait: Union[bool, Tuple[int, float]] = True) -> None:
        """Connect to a device in a running network scenario, using the specified shell.

        Args:
            machine (Machine): The device to connect.
            shell (str): The name of the shell to use for connecting.
            logs (bool): If True, print startup logs on stdout.
            wait (Union[bool, Tuple[int, float]]): If True, wait indefinitely until the end of the startup commands
                execution before connecting. If a tuple is provided, the first value indicates the number of retries
                before stopping waiting and the second value indicates the time interval to wait for each retry.
                Default is True.

        Returns:
            None

        Raises:
            LabNotFoundError: If the specified device is not associated to any network scenario.
        """
        raise NotImplementedError("You must implement `connect_tty_obj` method.")

    @abstractmethod
    def exec(self, machine_name: str, command: Union[List[str], str], lab_hash: Optional[str] = None,
             lab_name: Optional[str] = None, lab: Optional[Lab] = None, wait: Union[bool, Tuple[int, float]] = False,
             stream: bool = True) -> Union[IExecStream, Tuple[bytes, bytes, int]]:
        """Exec a command on a device in a running network scenario.

        Args:
            machine_name (str): The name of the device to connect.
            command (Union[List[str], str]): The command to exec on the device.
            lab_hash (Optional[str]): The hash of the network scenario.
                Can be used as an alternative to lab_name and lab. If None, lab_name or lab should be set.
            lab_name (Optional[str]): The name of the network scenario.
                Can be used as an alternative to lab_hash and lab. If None, lab_hash or lab should be set.
            lab (Optional[Kathara.model.Lab]): The network scenario object.
                Can be used as an alternative to lab_hash and lab_name. If None, lab_hash or lab_name should be set.
            wait (Union[bool, Tuple[int, float]]): If True, wait indefinitely until the end of the startup commands
                execution before executing the command. If a tuple is provided, the first value indicates the
                number of retries before stopping waiting and the second value indicates the time interval to wait
                for each retry. Default is False.
            stream (bool): If True, return an IExecStream object. If False,
                returns a tuple containing the complete stdout, the stderr, and the return code of the command.

        Returns:
            Union[IExecStream, Tuple[bytes, bytes, int]]: An IExecStream object or
            a tuple containing the stdout, the stderr and the return code of the command.

        Raises:
            InvocationError: If a running network scenario hash or name is not specified.
            MachineNotRunningError: If the specified device is not running.
            MachineBinaryError: If the binary of the command is not found.
        """
        raise NotImplementedError("You must implement `exec` method.")

    @abstractmethod
    def exec_obj(self, machine: Machine, command: Union[List[str], str], wait: Union[bool, Tuple[int, float]] = False,
                 stream: bool = True) -> Union[IExecStream, Tuple[bytes, bytes, int]]:
        """Exec a command on a device in a running network scenario.

        Args:
            machine (Machine): The device to connect.
            command (Union[List[str], str]): The command to exec on the device.
            wait (Union[bool, Tuple[int, float]]): If True, wait indefinitely until the end of the startup commands
                execution before executing the command. If a tuple is provided, the first value indicates the
                number of retries before stopping waiting and the second value indicates the time interval to wait
                for each retry. Default is False.
            stream (bool): If True, return an IExecStream object. If False,
                returns a tuple containing the complete stdout, the stderr, and the return code of the command.

        Returns:
            Union[IExecStream, Tuple[bytes, bytes, int]]: An IExecStream object or
            a tuple containing the stdout, the stderr and the return code of the command.

        Raises:
            LabNotFoundError: If the specified device is not associated to any network scenario.
            MachineNotRunningError: If the specified device is not running.
            MachineBinaryError: If the binary of the command is not found.
        """
        raise NotImplementedError("You must implement `exec_obj` method.")

    @abstractmethod
    def copy_files(self, machine: Machine, guest_to_host: Dict[str, Union[str, io.IOBase]]) -> None:
        """Copy files on a running device in the specified paths.

        Args:
            machine (Kathara.model.Machine): A running device object. It must have the api_object field populated.
            guest_to_host (Dict[str, Union[str, io.IOBase]]): A dict containing the device path as key and a
                fileobj to copy in path as value or a path to a file.

        Returns:
            None
        """
        raise NotImplementedError("You must implement `copy_files` method.")

    @abstractmethod
    def retrieve_files(self, machine: Machine, src: str, dst: str) -> None:
        """Copy files from a running device path to the host.

        Args:
            machine (Kathara.model.Machine): A running device object. It must have the api_object field populated.
            src (str): The path of the file or folder to copy from the device.
            dst (str): The destination path on the host.

        Returns:
            None
        """
        raise NotImplementedError("You must implement `retrieve_files` method.")

    @abstractmethod
    def get_machine_api_object(self, machine_name: str, lab_hash: Optional[str] = None, lab_name: Optional[str] = None,
                               lab: Optional[Lab] = None, all_users: bool = False) -> Any:
        """Return the corresponding API object of a running device in a network scenario.

        Args:
            machine_name (str): The name of the device.
            lab_hash (Optional[str]): The hash of the network scenario.
                Can be used as an alternative to lab_name and lab. If None, lab_name or lab should be set.
            lab_name (Optional[str]): The name of the network scenario.
                Can be used as an alternative to lab_hash and lab. If None, lab_hash or lab should be set.
            lab (Optional[Kathara.model.Lab]): The network scenario object.
                Can be used as an alternative to lab_hash and lab_name. If None, lab_hash or lab_name should be set.
            all_users (bool): If True, return information about devices of all users.

        Returns:
            Any: API object of the device specific for the current manager.

        Raises:
            InvocationError: If a running network scenario hash or name is not specified.
            MachineNotFoundError: If the specified device is not found.
        """
        raise NotImplementedError("You must implement `get_machine_api_object` method.")

    @abstractmethod
    def get_machines_api_objects(self, lab_hash: Optional[str] = None, lab_name: Optional[str] = None,
                                 lab: Optional[Lab] = None, all_users: bool = False) -> List[Any]:
        """Return API objects of running devices.

        Args:
            lab_hash (Optional[str]): The hash of the network scenario.
                Can be used as an alternative to lab_name and lab.
            lab_name (Optional[str]): The name of the network scenario.
                Can be used as an alternative to lab_hash and lab.
            lab (Optional[Kathara.model.Lab]): The network scenario object.
                Can be used as an alternative to lab_hash and lab_name.
            all_users (bool): If True, return information about devices of all users.

        Returns:
            List[Any]: API objects of devices, specific for the current manager.
        """
        raise NotImplementedError("You must implement `get_machines_api_objects` method.")

    @abstractmethod
    def get_link_api_object(self, link_name: str, lab_hash: Optional[str] = None, lab_name: Optional[str] = None,
                            lab: Optional[Lab] = None, all_users: bool = False) -> Any:
        """Return the corresponding API object of a collision domain in a network scenario.

        Args:
            link_name (str): The name of the collision domain.
            lab_hash (Optional[str]): The hash of the network scenario.
                Can be used as an alternative to lab_name and lab. If None, lab_name or lab should be set.
            lab_name (Optional[str]): The name of the network scenario.
                Can be used as an alternative to lab_hash and lab. If None, lab_hash or lab should be set.
            lab (Optional[Kathara.model.Lab]): The network scenario object.
                Can be used as an alternative to lab_hash and lab_name. If None, lab_hash or lab_name should be set.
            all_users (bool): If True, return information about collision domains of all users.

        Returns:
            Any: API object of the collision domain specific for the current manager.

         Raises:
            InvocationError: If a running network scenario hash or name is not specified.
            LinkNotFoundError: If the collision domain is not found.
        """
        raise NotImplementedError("You must implement `get_link_api_object` method.")

    @abstractmethod
    def get_links_api_objects(self, lab_hash: Optional[str] = None, lab_name: Optional[str] = None,
                              lab: Optional[Lab] = None, all_users: bool = False) -> List[Any]:
        """Return API objects of collision domains in a network scenario.

        Args:
            lab_hash (Optional[str]): The hash of the network scenario.
                Can be used as an alternative to lab_name and lab.
            lab_name (Optional[str]): The name of the network scenario.
                Can be used as an alternative to lab_hash and lab.
            lab (Optional[Kathara.model.Lab]): The network scenario object.
                Can be used as an alternative to lab_hash and lab_name.
            all_users (bool): If True, return information about collision domains of all users.

        Returns:
            List[Any]: API objects of collision domains, specific for the current manager.
        """
        raise NotImplementedError("You must implement `get_links_api_objects` method.")

    @abstractmethod
    def get_lab_from_api(self, lab_hash: Optional[str] = None, lab_name: Optional[str] = None) -> Lab:
        """Return the network scenario (specified by the hash or name), building it from API objects.

        Args:
            lab_hash (Optional[str]): The hash of the network scenario.
                Can be used as an alternative to lab_name. If None, lab_name should be set.
            lab_name (Optional[str]): The name of the network scenario.
                Can be used as an alternative to lab_hash. If None, lab_hash should be set.

        Returns:
            Lab: The built network scenario.

        Raises:
            InvocationError: If a running network scenario hash or name is not specified.
        """
        raise NotImplementedError("You must implement `get_lab_from_api` method.")

    @abstractmethod
    def update_lab_from_api(self, lab: Lab) -> None:
        """Update the passed network scenario from API objects.

        Args:
            lab (Lab): The network scenario to update.
        """
        raise NotImplementedError("You must implement `update_lab_from_api` method.")

    @abstractmethod
    def get_machines_stats(self, lab_hash: Optional[str] = None, lab_name: Optional[str] = None,
                           lab: Optional[Lab] = None, machine_name: str = None, all_users: bool = False) \
            -> Generator[Dict[str, IMachineStats], None, None]:
        """Return information about the running devices.

        Args:
            lab_hash (Optional[str]): The hash of the network scenario.
                Can be used as an alternative to lab_name and lab.
            lab_name (Optional[str]): The name of the network scenario.
                Can be used as an alternative to lab_hash and lab.
            lab (Optional[Kathara.model.Lab]): The network scenario object.
                Can be used as an alternative to lab_hash and lab_name.
            machine_name (str): If specified return all the devices with machine_name.
            all_users (bool): If True, return information about the device of all users.

        Returns:
              Generator[Dict[str, IMachineStats], None, None]: A generator containing dicts that has API Object
              identifier as keys and IMachineStats objects as values.
        """
        raise NotImplementedError("You must implement `get_machines_stats` method.")

    @abstractmethod
    def get_machine_stats(self, machine_name: str, lab_hash: Optional[str] = None, lab_name: Optional[str] = None,
                          lab: Optional[Lab] = None, all_users: bool = False) \
            -> Generator[Optional[IMachineStats], None, None]:
        """Return information of the specified device in a specified network scenario.

        Args:
            machine_name (str): The name of the device for which statistics are requested.
            lab_hash (Optional[str]): The hash of the network scenario.
                Can be used as an alternative to lab_name and lab. If None, lab_name or lab should be set.
            lab_name (Optional[str]): The name of the network scenario.
                Can be used as an alternative to lab_hash and lab. If None, lab_hash or lab should be set.
            lab (Optional[Kathara.model.Lab]): The network scenario object.
                Can be used as an alternative to lab_hash and lab_name. If None, lab_hash or lab_name should be set.
            all_users (bool): If True, search the device among all the users devices.

        Returns:
            Generator[Optional[IMachineStats], None, None]: A generator containing the IMachineStats object
            with the device info. Returns None if the device is not found.

        Raises:
            InvocationError: If a running network scenario hash or name is not specified.
        """
        raise NotImplementedError("You must implement `get_machine_stats` method.")

    @abstractmethod
    def get_machine_stats_obj(self, machine: Machine, all_users: bool = False) \
            -> Generator[Optional[IMachineStats], None, None]:
        """Return information of the specified device in a specified network scenario.

        Args:
            machine (Machine): The device for which statistics are requested.
            all_users (bool): If True, search the device among all the users devices.

        Returns:
            Generator[Optional[IMachineStats], None, None]: A generator containing the IMachineStats object
            with the device info. Returns None if the device is not found.

        Raises:
            LabNotFoundError: If the specified device is not associated to any network scenario.
            MachineNotRunningError: If the specified device is not running.
        """
        raise NotImplementedError("You must implement `get_machine_stats_obj` method.")

    @abstractmethod
    def get_links_stats(self, lab_hash: Optional[str] = None, lab_name: Optional[str] = None, lab: Optional[Lab] = None,
                        link_name: str = None, all_users: bool = False) -> Generator[Dict[str, ILinkStats], None, None]:
        """Return information about deployed networks.

        Args:
            lab_hash (Optional[str]): The hash of the network scenario.
                Can be used as an alternative to lab_name and lab.
            lab_name (Optional[str]): The name of the network scenario.
                Can be used as an alternative to lab_hash and lab.
            lab (Optional[Kathara.model.Lab]): The network scenario object.
                Can be used as an alternative to lab_hash and lab_name.
           link_name (str): If specified return all the networks with link_name.
           all_users (bool): If True, return information about the networks of all users.

        Returns:
             Generator[Dict[str, ILinkStats], None, None]: A generator containing dicts that has API Object
             identifier as keys and ILinksStats objects as values.
        """
        raise NotImplementedError("You must implement `get_links_stats` method.")

    @abstractmethod
    def get_link_stats(self, link_name: str, lab_hash: Optional[str] = None, lab_name: Optional[str] = None,
                       lab: Optional[Lab] = None, all_users: bool = False) \
            -> Generator[Optional[ILinkStats], None, None]:
        """Return information of the specified deployed network in a specified network scenario.

        Args:
            link_name (str): The name of the collision domain for which statistics are requested.
            lab_hash (Optional[str]): The hash of the network scenario.
                Can be used as an alternative to lab_name and lab. If None, lab_name or lab should be set.
            lab_name (Optional[str]): The name of the network scenario.
                Can be used as an alternative to lab_hash and lab. If None, lab_hash or lab should be set.
            lab (Optional[Kathara.model.Lab]): The network scenario object.
                Can be used as an alternative to lab_hash and lab_name. If None, lab_hash or lab_name should be set.
            all_users (bool): If True, return information about the networks of all users.

        Returns:
            Generator[Optional[ILinkStats], None, None]: A generator containing the ILinkStats object
            with the network info. Returns None if the network is not found.

        Raises:
            InvocationError: If a running network scenario hash or name is not specified.
        """
        raise NotImplementedError("You must implement `get_link_stats` method.")

    @abstractmethod
    def get_link_stats_obj(self, link: Link, all_users: bool = False) -> Generator[Optional[ILinkStats], None, None]:
        """Return information of the specified deployed network in a specified network scenario.

        Args:
            link (Link): The collision domain for which statistics are requested.
            all_users (bool): If True, return information about the networks of all users.

        Returns:
            Generator[Optional[ILinkStats], None, None]: A generator containing the ILinkStats object
            with the network info. Returns None if the network is not found.

        Raises:
            LabNotFoundError: If the specified device is not associated to any network scenario.
        """
        raise NotImplementedError("You must implement `get_link_stats_obj` method.")

    @abstractmethod
    def check_image(self, image_name: str) -> None:
        """Check if the specified image is valid.

        Args:
            image_name (str): The name of the image

        Returns:
            None

        Raises:
            ConnectionError: If the image is not locally available and there is no connection to a remote image repository.
            ImageNotFoundError: If the image is not found.
        """
        raise NotImplementedError("You must implement `check_image` method.")

    @abstractmethod
    def get_release_version(self) -> str:
        """Return the current manager version.

        Returns:
            str: The current manager version.
        """
        raise NotImplementedError("You must implement `get_release_version` method.")

    @staticmethod
    def get_formatted_manager_name() -> str:
        """Return a formatted string containing the current manager name.

        Returns:
            str: A formatted string containing the current manager name.
        """
        raise NotImplementedError("You must implement `get_formatted_manager_name` method.")



from ....factory.Factory import Factory


class OptionsHandlerFactory(Factory):
    def __init__(self) -> None:
        self.module_template: str = "Kathara.cli.ui.setting"
        self.name_template: str = "%sOptionsHandler"



from typing import Any

from docker import DockerClient

from .session.DockerTTYTerminalSession import DockerTTYTerminalSession
from ....foundation.manager.terminal.console.UnixConsoleAdapter import UnixConsoleAdapter
from ....foundation.manager.terminal.core.IConsoleAdapter import IConsoleAdapter
from ....foundation.manager.terminal.core.ITerminalSession import ITerminalSession
from ....foundation.manager.terminal.core.TerminalRunner import TerminalRunner


class DockerTTYTerminal(object):
    """High-level terminal runner for Docker over TTY sessions on Unix platforms.

    Args:
        handler (Any): The exec session handler.
        client (DockerClient): Docker client instance.
        exec_id (int): Docker exec ID identifying the running exec session.
    """

    __slots__ = ["_handler", "_runner"]

    def __init__(self, handler: Any, client: DockerClient, exec_id: str) -> None:
        console: IConsoleAdapter = UnixConsoleAdapter()
        session: ITerminalSession = DockerTTYTerminalSession(handler=handler, client=client, exec_id=exec_id)

        self._handler: Any = handler
        self._runner: TerminalRunner = TerminalRunner(console=console, session=session)

    def start(self) -> None:
        """Start the interactive terminal session.

        Returns:
            None
        """
        self._runner.start()

        # Force close the underlying object
        try:
            self._handler._response.close()
        except Exception:
            pass



import logging

import six
from abc import ABCMeta, abstractmethod


class InvalidValidator(Exception):
    """
    Raised when expected a valid validator but something else given
    """
    pass


@six.add_metaclass(ABCMeta)
class BaseValidator(object):
    """
    Validator Base class, each validator should inherit from this one
    """

    def __init__(self):
        logging.basicConfig()
        self.log = logging.getLogger(type(self).__name__)

    @abstractmethod
    def validate(self, input_string):
        """

        This function should be implemented in the validators

        :param input_string: Input string from command line (provided by the user)
        :return: True in case validation success / False otherwise
        """
        pass



import math
import sys
from itertools import islice
from multiprocessing import cpu_count
from sys import platform as _platform

import base64
import hashlib
import importlib
import io
import logging
import os
import re
import shutil
import tarfile
import tempfile
import unicodedata

from binaryornot.check import is_binary
from io import BytesIO
from platform import node, machine

from types import ModuleType
from typing import Any, Optional, Match, Generator, List, Callable, Union, Dict, Iterable, Tuple

from .exceptions import HostArchitectureError, InvocationError

# Platforms constants definition.
MAC_OS: str = "darwin"
WINDOWS: str = "win32"
LINUX: str = "linux"
LINUX2: str = "linux2"

# List of ignored files
EXCLUDED_FILES: List[str] = ['.DS_Store']

# Reserved names for devices
RESERVED_MACHINE_NAMES: List[str] = ['shared', '_test']


# Generic Functions
def check_python_version() -> None:
    if sys.version_info < (3, 9):
        logging.critical("Python version should be at least 3.9")
        sys.exit(1)


def class_for_name(module_name: str, class_name: str) -> Any:
    m = importlib.import_module(module_name + "." + class_name)
    return getattr(m, class_name)


def generate_urlsafe_hash(string: str) -> str:
    string = re.sub(r'[^\x00-\x7F]+', '', string)
    return base64.urlsafe_b64encode(hashlib.md5(string.encode('utf-8', errors='ignore')).digest())[:-2] \
        .decode('utf-8') \
        .replace('-', '').replace('_', '')


def get_absolute_path(path: str) -> str:
    abs_path = os.path.realpath(path)
    return abs_path if not os.path.islink(abs_path) else os.readlink(abs_path)


def get_executable_path(exec_path: str) -> Optional[str]:
    exec_abs_path = get_absolute_path(exec_path)

    if os.path.exists(exec_abs_path) and os.path.isfile(exec_abs_path):
        # If kathara is launched as a python script
        exec_abs_path = "\"" + exec_abs_path + "\""
        if exec_path.endswith("kathara.py"):
            # Prepend python in windows because it has no shebang
            return exec_by_platform(lambda: exec_abs_path,
                                    lambda: "%s %s" % (sys.executable, exec_abs_path),
                                    lambda: "%s %s" % (sys.executable, exec_abs_path)
                                    )
        else:
            # Maybe the executable is not in path, but is still a binary file
            return exec_abs_path
    else:
        # If kathara is in the path, return the absolute path of kathara
        which_path = shutil.which(exec_path)
        if which_path:
            return "\"" + which_path + "\""

    return None


def re_search_fail(expression: str, line: str) -> Match:
    matches = re.search(expression, line)

    if not matches:
        raise ValueError()

    return matches


def list_chunks(iterable: List, size: int) -> Generator:
    it = iter(iterable)
    item = list(islice(it, size))
    while item:
        yield item
        item = list(islice(it, size))


def chunk_list(iterable: Union[List, Iterable], size: int) -> List[List]:
    return [iterable] if len(iterable) < size else list_chunks(iterable, size)


def get_pool_size() -> int:
    return cpu_count()


def check_single_not_none_var(**kwargs) -> None:
    not_none = [x for x in kwargs.values() if x is not None]

    if len(not_none) > 1:
        raise InvocationError(f"You must specify only a parameter among {', '.join(kwargs.keys())}")


def check_required_single_not_none_var(**kwargs) -> None:
    not_none = [x for x in kwargs.values() if x is not None]

    if len(not_none) == 0:
        raise InvocationError(f"You must specify a parameter among {', '.join(kwargs.keys())}")
    elif len(not_none) > 1:
        raise InvocationError(f"You must specify only a parameter among {', '.join(kwargs.keys())}")


# Platform Specific Functions
def is_platform(desired_platform: str) -> bool:
    return _platform == desired_platform


def is_wsl_platform() -> bool:
    if _platform == LINUX or _platform == LINUX2:
        info = os.uname()
        return 'microsoft' in info.release.lower()
    return False


def exec_by_platform(fun_linux: Callable, fun_windows: Callable, fun_mac: Callable) -> Any:
    if _platform == LINUX or _platform == LINUX2:
        return fun_linux()
    elif _platform == WINDOWS:
        return fun_windows()
    elif _platform == MAC_OS:
        return fun_mac()


def pywintypes_import_stub() -> ModuleType:
    """Stub module of pywintypes for Unix systems (so it won't raise any `module not found` exception)."""
    import types
    from requests.exceptions import ConnectionError as RequestsConnectionError
    pywintypes = types.ModuleType("pywintypes")
    pywintypes.error = RequestsConnectionError
    return pywintypes


def pywintypes_import_win() -> ModuleType:
    import pywintypes
    return pywintypes


def import_pywintypes() -> ModuleType:
    return exec_by_platform(pywintypes_import_stub, pywintypes_import_win, pywintypes_import_stub)


def wait_user_input_linux() -> list:
    """Non-blocking input function for Linux and macOS."""
    import select
    to_break, _, _ = select.select([sys.stdin], [], [], 0.1)
    return to_break


def wait_user_input_windows() -> bool:
    """Return True if an Enter keypress is waiting to be read. Only for Windows."""
    import msvcrt
    return b'\r' in msvcrt.getch() if msvcrt.kbhit() else False


def convert_win_2_linux(filename: str, write: bool = False) -> Optional[bytes]:
    if not is_binary(filename):
        file_obj = None
        try:
            file_obj = open(filename, mode='r', encoding='utf-8-sig')
            file_content = file_obj.read().replace("\n\r", "\n").replace("\r\n", "\n")
            if not write:
                return file_content.encode('utf-8')
            else:
                with open(filename, mode='w', encoding='utf-8', newline="\n") as file_obj_write:
                    file_obj_write.write(file_content)
                return
        except Exception:
            # In any case the binaryornot heuristics fail (so the file is not a text file), close the stream and
            # read the file as a standard binary file.
            if file_obj:
                file_obj.close()
            pass

    if not write:
        return open(filename, mode='rb').read()


def is_admin() -> bool:
    def unix_root():
        return os.getuid() == 0

    def windows_admin():
        import ctypes
        return ctypes.windll.shell32.IsUserAnAdmin() != 0

    return exec_by_platform(unix_root, windows_admin, unix_root)


def get_current_user_home() -> str:
    def passwd_home():
        user_info = get_current_user_info()
        return user_info.pw_dir

    def default_home():
        return os.path.expanduser('~')

    return exec_by_platform(passwd_home, default_home, default_home)


def get_current_user_uid_gid() -> (int, int):
    def unix():
        user_info = get_current_user_info()
        return user_info.pw_uid, user_info.pw_gid

    return exec_by_platform(unix, lambda: (None, None), unix)

def slug(value):
    value = unicodedata.normalize("NFKD", value).encode("ascii", "ignore").decode("ascii")
    value = re.sub(r"[^\w\s-]", "", value).strip().lower()
    return re.sub(r"[-\s]+", "-", value)

def get_current_user_name() -> str:
    hostname = generate_urlsafe_hash(node())

    def unix():
        user_info = get_current_user_info()
        return user_info.pw_name

    def windows():
        import getpass
        return getpass.getuser()

    return slug("%s-%s" % (exec_by_platform(unix, windows, unix), hostname))


def get_current_user_info() -> Any:
    def passwd_info():
        import pwd
        user_id = os.getuid()

        # It's root, take the real user from env variable
        if user_id == 0:
            # If it's a sudoer, take the SUDO_UID and use it as user_id variable
            # If not, keep using the user_id = 0
            real_user_id = os.environ.get("SUDO_UID")

            if real_user_id:
                user_id = int(real_user_id)

        return pwd.getpwuid(user_id)

    return exec_by_platform(passwd_info, lambda: None, passwd_info)


def check_directory_permissions(path: str, mode: str = "ro") -> list[str]:
    """
    Check directory permissions across different platforms.

    Args:
        path: Path to check permissions for
        mode: Permission mode to check ('r' for read, 'w' for write, 'x' for execute)

    Returns:
        list[str]: A list containing any missing permissions
    """

    def check_directory_permissions_unix() -> list[str]:
        if not os.path.exists(path):
            raise FileExistsError(f"Path `{path}` does not exist.")

        if not os.path.isdir(path):
            raise NotADirectoryError(f"Path `{path}` must be a directory.")

        missing = []

        if 'r' in mode and not os.access(path, os.R_OK):
            missing.append("read (r)")
        if 'w' in mode and not os.access(path, os.W_OK):
            missing.append("write (w)")
        if 'x' in mode and not os.access(path, os.X_OK):
            missing.append("execute (x)")

        return missing

    def check_directory_permissions_windows() -> list[str]:
        import ctypes
        from ctypes import wintypes

        if not os.path.exists(path):
            raise FileExistsError(f"Path `{path}` does not exist.")

        if not os.path.isdir(path):
            raise NotADirectoryError(f"Path `{path}` must be a directory.")

        missing = []

        generic_read = 0x80000000
        generic_write = 0x40000000
        generic_execute = 0x20000000

        kernel32 = ctypes.WinDLL('kernel32', use_last_error=True)

        create_file = kernel32.CreateFileW
        create_file.argtypes = [
            wintypes.LPCWSTR,
            wintypes.DWORD,
            wintypes.DWORD,
            wintypes.LPVOID,
            wintypes.DWORD,
            wintypes.DWORD,
            wintypes.HANDLE
        ]
        create_file.restype = wintypes.HANDLE

        open_existing = 3
        file_flag_backup_semantics = 0x02000000
        invalid_handle_value = wintypes.HANDLE(-1).value
        file_share_read = 0x00000001
        file_share_write = 0x00000002

        close_handle = kernel32.CloseHandle
        close_handle.argtypes = [wintypes.HANDLE]
        close_handle.restype = wintypes.BOOL

        if 'r' in mode:
            handle = create_file(
                path,
                generic_read,
                file_share_read | file_share_write,
                None,
                open_existing,
                file_flag_backup_semantics,
                None
            )

            if handle == invalid_handle_value:
                missing.append("read (r)")
            else:
                close_handle(handle)

        if 'w' in mode:
            handle = create_file(
                path,
                generic_write,
                file_share_read | file_share_write,
                None,
                open_existing,
                file_flag_backup_semantics,
                None
            )

            if handle == invalid_handle_value:
                missing.append("write (w)")
            else:
                close_handle(handle)

        if 'x' in mode:
            handle = create_file(
                path,
                generic_execute,
                file_share_read | file_share_write,
                None,
                open_existing,
                file_flag_backup_semantics,
                None
            )

            if handle == invalid_handle_value:
                missing.append("execute (x)")
            else:
                close_handle(handle)

        return missing

    return exec_by_platform(
        check_directory_permissions_unix,
        check_directory_permissions_windows,
        check_directory_permissions_unix
    )


# Architecture Test
def get_architecture() -> str:
    architecture = machine().lower()

    logging.debug("Machine architecture is `%s`." % architecture)

    if architecture in ["x86_64", "amd64"]:
        return "amd64"
    elif architecture == "i686":
        return "386"
    elif architecture in ["arm64", "aarch64"]:
        return "arm64"
    elif architecture == "armv7l":
        return "armv7"
    elif architecture == "armv6l":
        return "armv6"
    else:
        raise HostArchitectureError(architecture)


# Formatting Functions
def human_readable_bytes(size_bytes: int) -> str:
    if size_bytes == 0:
        return "0 B"

    size_name = ("B", "KB", "MB", "GB", "TB", "PB", "EB", "ZB", "YB")
    i = int(math.floor(math.log(size_bytes, 1024)))
    p = math.pow(1024, i)
    s = round(size_bytes / p, 2)

    return "%s %s" % (s, size_name[i])


# Lab Functions
def pack_file_for_tar(file_obj: Union[str, io.IOBase], arc_name: str) -> (tarfile.TarInfo, bytes):
    if isinstance(file_obj, str):
        file_content_patched = convert_win_2_linux(file_obj)
        file_content = BytesIO(file_content_patched)
        filesize = len(file_content_patched)
    elif isinstance(file_obj, io.IOBase):
        file_content = file_obj.read()
        file_content = BytesIO(file_content.encode('utf-8') if isinstance(file_obj, io.TextIOBase) else file_content)
        file_obj.seek(0, 2)
        filesize = file_obj.tell()
        file_obj.seek(0)
    else:
        raise ValueError("File type %s not supported" % type(file_obj))

    tarinfo = tarfile.TarInfo(arc_name.replace("\\", "/"))  # Tar files must have Linux-style paths
    tarinfo.size = filesize

    return tarinfo, file_content


def pack_files_for_tar(guest_to_host: Dict) -> bytes:
    with tempfile.NamedTemporaryFile(mode='wb+', suffix='.tar.gz') as temp_file:
        with tarfile.open(fileobj=temp_file, mode='w:gz') as tar_file:
            for path, file_obj in guest_to_host.items():
                tar_info, file_content = pack_file_for_tar(file_obj, arc_name=path)
                tar_file.addfile(tar_info, file_content)

        temp_file.seek(0)

        tar_data = temp_file.read()

    return tar_data


def parse_cd_mac_address(value) -> Tuple[str, str]:
    if '/' in value:
        parts = [x for x in value.split('/') if x]

        if len(parts) != 2:
            raise SyntaxError(f"Invalid interface definition: `{value}`.")

        cd_name = parts[0]
        mac_address = parts[1]
    else:
        (cd_name, mac_address) = value, None

    return cd_name, mac_address


# Docker Engine Utils
def parse_docker_engine_version(v: str) -> str:
    parts = []
    for part in v.split('.'):
        numeric_part = ""
        # Iterate over the single chars of the part
        for char in part:
            # As soon as we get a non-digit value, we break
            if char.isdigit():
                numeric_part += char
            else:
                break

        # If the numeric_part is not empty, append it as a chunk of the version
        if numeric_part:
            parts.append(numeric_part)
        else:
            break

    return '.'.join(parts)



from abc import ABC, abstractmethod
from typing import Any, Dict


class IMachineStats(ABC):
    """Interface for handling the statistics of deployed devices"""

    @abstractmethod
    def update(self) -> None:
        """Update dynamic statistics with the current ones.

        Returns:
            None
        """
        raise NotImplementedError("You must implement `update` method.")

    @abstractmethod
    def to_dict(self) -> Dict[str, Any]:
        """Transform statistics into a dict representation.

        Returns:
            Dict[str, Any]: Dict containing statistics.
        """
        raise NotImplementedError("You must implement `to_dict` method.")



from ...ui.utils import confirmation_prompt
from ....manager.docker.DockerImage import DockerImage
from ....setting.Setting import Setting


class UpdateDockerImage(object):
    """Listener fired when there is a Docker Image update."""

    def run(self, docker_image: DockerImage, image_name: str) -> None:
        """Prompt the user for Docker Image Update.

        Args:
            docker_image (DockerImage): DockerImage instance.
            image_name (str): Name of the Docker image to update.

        Returns:
            None
        """
        policy = Setting.get_instance().image_update_policy
        if policy == 'Prompt':
            confirmation_prompt("A new version of image `%s` has been found on Docker Hub. "
                                "Do you want to pull it?" % image_name,
                                lambda: docker_image.pull(image_name),
                                lambda: None
                                )
        elif policy == 'Always':
            docker_image.pull(image_name)



from . import ConsoleMenu
from .items import SelectionItem


class SelectionMenu(ConsoleMenu):
    """
    A menu that simplifies item creation, just give it a list of strings and it builds the menu for you

    Args:
        strings (:obj:`list` of :obj:`str`):  The list of strings this menu should be built from.
        title (str): The title of the menu.
        subtitle (str, lambda): The subtitle of the menu.
        screen (:obj:`consolemenu.screen.Screen`): The screen object associated with this menu.
        formatter (:obj:`MenuFormatBuilder`): The MenuFormatBuilder instance used to format this menu.
        prologue_text (str): Text to include in the "prologue" section of the menu.
        epilogue_text (str): Text to include in the "epilogue" section of the menu.
        show_exit_option (bool): Specifies whether this menu should show an exit item by default. Defaults to True.
            Can be overridden when the menu is started.
        exit_option_text (str): Text for the Exit menu item. Defaults to 'Exit'.
    """

    def __init__(self, strings, title=None, subtitle=None, screen=None, formatter=None,
                 prologue_text=None, epilogue_text=None, show_exit_option=True, exit_option_text='Exit'):
        super(SelectionMenu, self).__init__(title, subtitle, screen=screen, formatter=formatter,
                                            prologue_text=prologue_text, epilogue_text=epilogue_text,
                                            show_exit_option=show_exit_option, exit_option_text=exit_option_text)
        for index, item in enumerate(strings):
            self.append_item(SelectionItem(item, index, self))

    @classmethod
    def get_selection(cls, strings, title="Select an option", subtitle=None, show_exit_option=True, _menu=None):
        """
        Single-method way of getting a selection out of a list of strings.

        Args:
            strings (:obj:`list` of :obj:`str`):  The list of strings this menu should be built from.
            title (str): The title of the menu.
            subtitle (str): The subtitle of the menu.
            show_exit_option (bool): Specifies whether this menu should show an exit item by default. Defaults to True.
            _menu: Should probably only be used for testing, pass in a list and the created menu used internally by
                the method will be appended to it

        Returns:
            int: The index of the selected option.

        """
        menu = cls(strings, title, subtitle, show_exit_option=show_exit_option)
        if _menu is not None:
            _menu.append(menu)
        menu.show()
        menu.join()
        return menu.selected_option

    def append_string(self, string):
        self.append_item(SelectionItem(string))



import argparse
import logging
import re
import shlex
import subprocess
import sys
from datetime import datetime
from typing import Any, Dict, Generator, Optional, Union, Tuple
from typing import Callable

from rich import box
from rich.console import RenderableType, Group
from rich.highlighter import RegexHighlighter
from rich.panel import Panel
from rich.prompt import Confirm
from rich.table import Table
from rich.text import Text

from ... import utils
from ...foundation.manager.stats.IMachineStats import IMachineStats
from ...model.Lab import Lab
from ...setting.Setting import Setting
from ...utils import parse_cd_mac_address

FORBIDDEN_TABLE_COLUMNS = ["container_name"]


class LabMetaHighlighter(RegexHighlighter):
    """Highlights metadata of the network scenario."""
    base_style = "kathara."
    highlights = [
        re.compile(r"(?P<lab_name>^Name: )", re.MULTILINE),
        re.compile(r"(?P<lab_description>^Description: )", re.MULTILINE),
        re.compile(r"(?P<lab_version>^Version: )", re.MULTILINE),
        re.compile(r"(?P<lab_author>^Author\(s\): )", re.MULTILINE),
        re.compile(r"(?P<lab_email>^Email: )", re.MULTILINE),
        re.compile(r"(?P<lab_web>^Website: )", re.MULTILINE),
    ]


def confirmation_prompt(prompt_string: str, callback_yes: Callable, callback_no: Callable) -> Any:
    answer = Confirm.ask(prompt_string)
    if not answer:
        return callback_no()

    return callback_yes()


def create_panel(message: Union[str, Text], **kwargs) -> Panel:
    return Panel(
        Text.from_markup(
            message,
            style=kwargs['style'] if 'style' in kwargs else "none",
            justify=kwargs['justify'] if 'justify' in kwargs else None,
        ) if isinstance(message, str) else message,
        title=kwargs['title'] if 'title' in kwargs else None,
        title_align="center",
        box=kwargs['box'] if 'box' in kwargs else box.SQUARE,
    )


def create_lab_table(streams: Generator[Dict[str, IMachineStats], None, None]) -> Optional[RenderableType]:
    try:
        result = next(streams)
    except StopIteration:
        return None

    ts_header = f"TIMESTAMP: {datetime.now()}"
    if not result:
        return Group(
            Text(ts_header, style="italic", justify="center"),
            create_panel("No Devices Found", style="red bold", justify="center", box=box.DOUBLE)
        )

    table = Table(title=ts_header, show_lines=True, expand=True, box=box.SQUARE_DOUBLE_HEAD)

    for item in result.values():
        row_data = item.to_dict()
        row_data = dict(filter(lambda x: x[0] not in FORBIDDEN_TABLE_COLUMNS, row_data.items()))

        if not table.columns:
            for col in map(lambda x: x.replace('_', ' ').upper(), row_data.keys()):
                table.add_column(col, header_style="dark_orange3")

        table.add_row(*map(lambda x: str(x), row_data.values()))

    return table


def create_topology_table(lab: Lab) -> Optional[RenderableType]:
    ts_header = f"TIMESTAMP: {datetime.now()}"

    table = Table(title=ts_header, show_lines=True, box=box.SQUARE_DOUBLE_HEAD)

    if not lab.links:
        return Group(
            Text(ts_header, style="italic", justify="center"),
            create_panel("No Collision Domains Found", style="red bold", justify="center", box=box.DOUBLE)
        )

    for link in sorted(lab.links.values(), key=lambda x: x.name):
        row_data = {
            'LINK NAME': link.name,
            'DEVICES': ", ".join(link.machines.keys())
        }

        if not table.columns:
            for col in row_data.keys():
                table.add_column(col, header_style="dark_orange3")

        table.add_row(*map(lambda x: str(x), row_data.values()))

    return table


def open_machine_terminal(machine) -> None:
    """Connect to the device with the terminal specified in the settings.

    Returns:
        None
    """
    Setting.get_instance().check_terminal()
    terminal = Setting.get_instance().terminal

    logging.debug("Opening terminal for device %s.", machine.name)

    executable_path = utils.get_executable_path(sys.argv[0])

    if not executable_path:
        raise FileNotFoundError("Unable to find Kathara.")

    is_vmachine = "-v" if not machine.lab.has_host_path() else ""
    connect_command = f"\"{sys.executable}\" {executable_path} connect {is_vmachine} -l {machine.name}"

    logging.debug("Terminal will open in directory %s." % machine.lab.fs_path())

    def unix_connect() -> None:
        if terminal == "TMUX":
            from ...trdparty.libtmux.tmux import TMUX

            logging.debug("Attaching `%s` to TMUX session `%s` with command `%s`" % (machine.name, machine.lab.name,
                                                                                     connect_command))

            TMUX.get_instance().add_window(
                machine.lab.name,
                machine.name,
                connect_command,
                cwd=machine.lab.fs_path()
            )
        else:
            logging.debug("Opening Linux terminal with command: %s." % connect_command)

            # Command should be passed as an array
            # https://stackoverflow.com/questions/9935151/popen-error-errno-2-no-such-file-or-directory/9935511
            command = [terminal]
            if 'gnome-terminal' in terminal:
                command.append("--")
                command.extend(shlex.split(connect_command))
            else:
                command.append("-e")
                command.append(connect_command)
            subprocess.Popen(command,
                             cwd=machine.lab.fs_path(),
                             start_new_session=True
                             )

    def windows_connect() -> None:
        complete_win_command = "& %s" % connect_command
        logging.debug("Opening Windows terminal with command: %s." % complete_win_command)
        subprocess.Popen(["powershell.exe",
                          '-Command',
                          complete_win_command
                          ],
                         creationflags=subprocess.CREATE_NEW_CONSOLE,
                         cwd=machine.lab.fs_path()
                         )

    def osx_connect() -> None:
        cd_to_lab_path = "cd \"%s\" &&" % machine.lab.fs_path() if machine.lab.has_host_path() else ""
        complete_osx_command = "%s clear && %s && exit" % (cd_to_lab_path, connect_command)

        if terminal == "TMUX":
            from ...trdparty.libtmux.tmux import TMUX

            logging.debug("Attaching `%s` to TMUX session `%s` with command `%s`" % (machine.name, machine.lab.name,
                                                                                     complete_osx_command))

            TMUX.get_instance().add_window(
                machine.lab.name,
                machine.name,
                complete_osx_command,
                cwd=machine.lab.fs_path()
            )
        else:
            import appscript
            logging.debug("Opening OSX terminal with command: %s." % complete_osx_command)
            terminal_app = appscript.app(terminal)
            if terminal == 'iTerm':
                window = terminal_app.create_window_with_default_profile()
                window.current_session.write(text=complete_osx_command)
            elif terminal == 'Terminal':
                terminal_app.do_script(complete_osx_command)

    utils.exec_by_platform(unix_connect, windows_connect, osx_connect)


# Types for argparse
def alphanumeric(value: str, pat: re.Pattern = re.compile(r"^\w+$")) -> str:
    if not pat.match(value):
        raise argparse.ArgumentTypeError("invalid alphanumeric value")

    return value


def interface_cd_mac(value: str) -> Tuple[str, str, str]:
    n, cd, mac = None, None, None
    try:
        parts = value.split('/')
        (n, cd) = parts[0].split(':')
        if len(parts) == 2:
            if parts[1]:
                mac = parts[1]
            else:
                raise ValueError
    except ValueError:
        raise argparse.ArgumentTypeError("invalid interface definition: %s" % value)

    if not re.search(r"^\w+$", cd):
        raise argparse.ArgumentTypeError(f"invalid interface definition, "
                                         f"collision domain `{cd}` contains non-alphanumeric characters")

    return n, cd, mac


def cd_mac(value) -> Tuple[str, str]:
    return parse_cd_mac_address(value)


def volume(value: str) -> str:
    values = list(filter(lambda x: x, value.split('|')))
    if len(values) != 3 and len(values) != 2:
        raise argparse.ArgumentTypeError("invalid volume definition: %s" % value)

    return value



import base64
import json
import os

from . import utils as setting_utils
from ....foundation.cli.ui.setting.OptionsHandler import OptionsHandler
from ....setting.addon.KubernetesSettingsAddon import DEFAULTS, DEFAULT_DOCKER_CONFIG_JSON_PATH
from ....trdparty.consolemenu import *
from ....trdparty.consolemenu.items import *
from ....trdparty.consolemenu.validators.regex import RegexValidator
from ....validator.DockerConfigJsonValidator import DockerConfigJsonValidator


class KubernetesOptionsHandler(OptionsHandler):
    def add_items(self, current_menu: ConsoleMenu, menu_formatter: MenuFormatBuilder) -> None:
        # API URL Option
        api_server_url_string = "Insert a Kubernetes API Server URL"
        api_server_url_menu = SelectionMenu(
            strings=[],
            title=api_server_url_string,
            subtitle=setting_utils.current_string("api_server_url"),
            prologue_text="""You can specify a remote Kubernetes API Server URL to """
                          """connect to when Megalos is not used on a Kubernetes master.
                          
                          Default is %s.""" % DEFAULTS['api_server_url'],
            formatter=menu_formatter
        )

        api_server_url_menu.append_item(
            FunctionItem(
                text=api_server_url_string,
                function=setting_utils.update_value,
                args=['api_server_url',
                      RegexValidator(setting_utils.URL_REGEX),
                      'Write a Kubernetes API Server URL:',
                      'Kubernetes API Server URL is not a valid URL (remove '
                      'the trailing slash, if present)'
                      ],
                should_exit=True
            )
        )
        api_server_url_menu.append_item(
            FunctionItem(
                text="Reset value to Empty String",
                function=setting_utils.update_setting_value,
                args=["api_server_url", None],
                should_exit=True
            )
        )

        api_url_item = SubmenuItem(api_server_url_string, api_server_url_menu, current_menu)

        # API Token Option
        api_token_string = "Insert a Kubernetes API Token"
        api_token_menu = SelectionMenu(
            strings=[],
            title=api_token_string,
            prologue_text="""When using a remote Kubernetes API Server, you must also """
                          """specify the authentication token to use.
                          
                          Default is %s.""" % DEFAULTS['api_token'],
            formatter=menu_formatter
        )

        api_token_menu.append_item(
            FunctionItem(
                text=api_token_string,
                function=setting_utils.update_value,
                args=['api_token',
                      RegexValidator(r'^.+$'),
                      'Write a Kubernetes API Token:',
                      'Kubernetes API Token not valid!'
                      ],
                should_exit=True
            )
        )
        api_token_menu.append_item(
            FunctionItem(
                text="Reset value to Empty String",
                function=setting_utils.update_setting_value,
                args=["api_token", None],
                should_exit=True
            )
        )

        api_token_item = SubmenuItem(api_token_string, api_token_menu, current_menu)

        # Shared Mount Option
        host_shared_string = "Automatically mount /shared on startup"
        host_shared_menu = SelectionMenu(
            strings=[],
            title=host_shared_string,
            subtitle=setting_utils.current_bool("host_shared"),
            prologue_text="""Each Kubernetes worker node creates a /home/shared """
                          """directory and it is made available for reading/writing """
                          """inside the device under the special directory `/shared`.
                          
                          Default is %s.""" %
                          setting_utils.format_bool(DEFAULTS['host_shared']),
            formatter=menu_formatter
        )

        host_shared_menu.append_item(
            FunctionItem(
                text="Yes",
                function=setting_utils.update_setting_value,
                args=["host_shared", True],
                should_exit=True
            )
        )
        host_shared_menu.append_item(
            FunctionItem(
                text="No",
                function=setting_utils.update_setting_value,
                args=["host_shared", False],
                should_exit=True
            )
        )

        host_shared_item = SubmenuItem(host_shared_string, host_shared_menu, current_menu)

        # Image Pull Policy Option
        image_pull_policy_string = "Image Pull Policy"
        image_pull_policy_menu = SelectionMenu(
            strings=[],
            title=image_pull_policy_string,
            subtitle=setting_utils.current_string("image_pull_policy"),
            prologue_text="""Specify the image pull policy for Docker images used by devices.
             
                          Default is `%s`.""" % DEFAULTS['image_pull_policy'],
            formatter=menu_formatter
        )

        image_pull_policy_menu.append_item(
            FunctionItem(
                text="Always",
                function=setting_utils.update_setting_value,
                args=["image_pull_policy", "Always"],
                should_exit=True
            )
        )
        image_pull_policy_menu.append_item(
            FunctionItem(
                text="If Not Present",
                function=setting_utils.update_setting_value,
                args=["image_pull_policy", "IfNotPresent"],
                should_exit=True
            )
        )
        image_pull_policy_menu.append_item(
            FunctionItem(
                text="Never",
                function=setting_utils.update_setting_value,
                args=["image_pull_policy", "Never"],
                should_exit=True
            )
        )

        image_pull_policy_item = SubmenuItem(image_pull_policy_string, image_pull_policy_menu, current_menu)

        # Private Registry docker_config_json Option
        docker_config_json_string = "Configure a Docker Private Registry"
        docker_config_json_menu = SelectionMenu(
            strings=[],
            title=docker_config_json_string,
            subtitle=setting_utils.current_string("docker_config_json"),
            prologue_text="""Insert the config.json file path containing the configuration to """
                          """access private container registries. The content will be stored as """
                          """a base64 string and it will be used to create a Kubernetes Secret """
                          """of type `kubernetes.io/dockerconfigjson`.   
                          
                          Default is %s.""" % DEFAULTS['docker_config_json'],
            formatter=menu_formatter
        )

        def store_b64_docker_json_callback(result: str) -> str:
            with open(os.path.expanduser(result), 'r') as docker_config_json_file:
                return base64.b64encode(json.dumps(json.load(docker_config_json_file)).encode()).decode()

        docker_config_json_menu.append_item(
            FunctionItem(
                text=docker_config_json_string,
                function=setting_utils.update_value_with_default,
                args=['docker_config_json',
                      DockerConfigJsonValidator(),
                      'Write the path to the Docker Config JSON file',
                      None,
                      DEFAULT_DOCKER_CONFIG_JSON_PATH,
                      store_b64_docker_json_callback
                      ],
                should_exit=True
            )
        )
        docker_config_json_menu.append_item(
            FunctionItem(
                text="Reset value to Empty String",
                function=setting_utils.update_setting_value,
                args=["docker_config_json", None],
                should_exit=True
            )
        )

        docker_config_json_item = SubmenuItem(docker_config_json_string, docker_config_json_menu, current_menu)

        current_menu.append_item(api_url_item)
        current_menu.append_item(api_token_item)
        current_menu.append_item(host_shared_item)
        current_menu.append_item(image_pull_policy_item)
        current_menu.append_item(docker_config_json_item)



from typing import Optional, Dict

from rich.progress import Progress, TextColumn, BarColumn, TaskID, TaskProgressColumn, TimeRemainingColumn, \
    SpinnerColumn


class HandleDockerImagePull(object):
    """Listener for handling a Docker pull progress bar."""
    __slots__ = ['progress_bar', 'tasks']

    def __init__(self) -> None:
        self.progress_bar: Optional[Progress] = None
        self.tasks: Dict[str, TaskID] = {}

    def init(self) -> None:
        """Initialize a progress bar to track Docker image pull.

        Returns:
            None
        """
        self.progress_bar = Progress(
            TextColumn("[progress.description]{task.description}", justify='left'),
            SpinnerColumn(),
            BarColumn(bar_width=None, complete_style="default", finished_style="green"),
            TaskProgressColumn(justify='right'),
            TimeRemainingColumn(),
            expand=True,
        )

        self.progress_bar.start()

    def update(self, progress: Dict) -> None:
        """Update with a new progress coming from the Docker APIs.

        Args:
            progress (Dict): Progress coming from the pull API.

        Returns:
            None
        """
        if self.progress_bar:
            completed = False
            if progress['status'] == 'Download complete':
                description = f'[Download Complete {progress["id"]}]'
                completed = True
            elif progress['status'] == 'Downloading':
                description = f'[bold][Downloading {progress["id"]}]'
            else:
                return

            layer_id = progress["id"]
            if layer_id not in self.tasks.keys():
                if completed:
                    self.tasks[layer_id] = self.progress_bar.add_task(
                        description, total=100, completed=100, fields={'layer_id': layer_id}
                    )
                else:
                    total = progress['progressDetail']['total'] if 'total' in progress['progressDetail'] else None
                    self.tasks[layer_id] = self.progress_bar.add_task(
                        description, total=total, fields={'layer_id': layer_id}
                    )
            else:
                if completed:
                    self.progress_bar.update(self.tasks[layer_id], description=description, total=100, completed=100)
                else:
                    current = progress['progressDetail']['current'] if 'current' in progress['progressDetail'] else None
                    self.progress_bar.update(self.tasks[layer_id], completed=current)

    def finish(self) -> None:
        """Closes the Docker pull progress bar.

        Returns:
            None
        """
        if self.progress_bar:
            for task in self.progress_bar.tasks:
                description = f"[Download Complete {task.fields['fields']['layer_id']}]" \
                    if 'fields' in task.fields and 'layer_id' in task.fields['fields'] else None
                self.progress_bar.update(task.id, description=description, completed=task.total, total=task.total)
            self.progress_bar.stop()

            self.tasks = {}
            self.progress_bar = None

    def unregister(self) -> None:
        """Method called when the event is unregistered

        Returns:
            None
        """
        self.finish()



from .format.menu_borders import MenuBorderStyle, MenuBorderStyleFactory
from .format.menu_style import MenuStyle
from .menu_component import Dimension, MenuHeader, MenuTextSection, MenuItemsSection, MenuFooter, MenuPrompt


class MenuFormatBuilder(object):
    """
    Builder class for generating the menu format.
    """

    def __init__(self, max_dimension=None):
        if max_dimension is None:
            max_dimension = Dimension(width=80, height=40)
        self.__max_dimension = max_dimension
        self.__border_style_factory = MenuBorderStyleFactory()
        self.__header = MenuHeader(menu_style=MenuStyle(), max_dimension=max_dimension)
        self.__prologue = MenuTextSection(menu_style=MenuStyle(), max_dimension=max_dimension)
        self.__items_section = MenuItemsSection(menu_style=MenuStyle(), max_dimension=max_dimension)
        self.__epilogue = MenuTextSection(menu_style=MenuStyle(), max_dimension=max_dimension)
        self.__footer = MenuFooter(menu_style=MenuStyle(), max_dimension=max_dimension)
        self.__prompt = MenuPrompt(menu_style=MenuStyle(), max_dimension=max_dimension)
        # Indent items deeper than other sections
        self.__items_section.style.padding.left = 3
        # Change default top border of prompt to 0, so it hugs the bottom of the menu
        self.__prompt.style.padding.top = 0

    # ============================================================
    # Margins and Border style applies to entire menu
    # ============================================================

    def set_border_style(self, border_style):
        """
        Set the border style using the specified MenuBorderStyle instance.
        :param border_style: the instance of MenuBorderStyle to use for border style formatting.
        """
        if not isinstance(border_style, MenuBorderStyle):
            raise TypeError('border_style must be type MenuBorderStyle')
        self.__header.style.border_style = border_style
        self.__prologue.style.border_style = border_style
        self.__items_section.style.border_style = border_style
        self.__epilogue.style.border_style = border_style
        self.__footer.style.border_style = border_style
        self.__prompt.style.border_style = border_style
        return self

    def set_border_style_type(self, border_style_type):
        """
        Set the border style using the specified border style type. The border style type should be an
        integer value recognized by the border style factory for this formatter instance.
        The built-in border style types are provided by the `MenuBorderStyleType` class, or custom
        border style types can be provided if using a custom border style factory.
        :param border_style_type: an integer value representing the border style type.
        """
        style = self.__border_style_factory.create_border(border_style_type)
        self.set_border_style(style)
        return self

    def set_border_style_factory(self, border_style_factory):
        """
        Set the instance of MenuBorderStyleFactory to use for generating border styles.
        Typically, this method will never need to be used, unless the default MenuBorderStyleFactory
        has been subclassed to provide custom border styles.
        :param border_style_factory: an instance of MenuBorderStyleFactory.
        """
        if not isinstance(border_style_factory, MenuBorderStyleFactory):
            raise TypeError('border_style_factory must be type MenuBorderStyleFactory')
        self.__border_style_factory = border_style_factory
        return self

    def set_bottom_margin(self, bottom_margin):
        """
        Set the bottom margin of the menu. This will determine the number of console lines appear between the
        bottom of the menu border and the menu input prompt.
        :param bottom_margin: an integer value
        """
        self.__footer.style.margins.bottom = bottom_margin
        return self

    def set_left_margin(self, left_margin):
        """
        Set the left margin of the menu.  This will determine the number of spaces between the left edge of the
        screen and the left menu border.
        :param left_margin: an integer value
        """
        self.__header.style.margins.left = left_margin
        self.__prologue.style.margins.left = left_margin
        self.__items_section.style.margins.left = left_margin
        self.__epilogue.style.margins.left = left_margin
        self.__footer.style.margins.left = left_margin
        self.__prompt.style.margins.left = left_margin
        return self

    def set_right_margin(self, right_margin):
        """
        Set the right margin of the menu.  This will determine the number of spaces between the right edge of the
        screen and the right menu border.
        :param right_margin: an integer value
        """
        self.__header.style.margins.right = right_margin
        self.__prologue.style.margins.right = right_margin
        self.__items_section.style.margins.right = right_margin
        self.__epilogue.style.margins.right = right_margin
        self.__footer.style.margins.right = right_margin
        self.__prompt.style.margins.right = right_margin
        return self

    def set_top_margin(self, top_margin):
        """
        Set the top margin of the menu.  This will determine the number of console lines between the top edge
        of the screen and the top menu border.
        :param top_margin: an integer value
        """
        self.__header.style.margins.top = top_margin
        return self

    # ============================================================
    # Header Settings
    # ============================================================

    def set_title_align(self, align='left'):
        self.__header.title_align = align
        return self

    def set_subtitle_align(self, align='left'):
        self.__header.subtitle_align = align
        return self

    def set_header_left_padding(self, x):
        self.__header.style.padding.left = x
        return self

    def set_header_right_padding(self, x):
        self.__header.style.padding.right = x
        return self

    def set_header_bottom_padding(self, x):
        self.__header.style.padding.bottom = x
        return self

    def set_header_top_padding(self, x):
        self.__header.style.padding.top = x
        return self

    def show_header_bottom_border(self, flag):
        self.__header.show_bottom_border = flag
        return self

    # ============================================================
    # Footer Settings
    # ============================================================

    def set_footer_left_padding(self, x):
        self.__footer.style.padding.left = x
        return self

    def set_footer_right_padding(self, x):
        self.__footer.style.padding.right = x
        return self

    def set_footer_bottom_padding(self, x):
        self.__footer.style.padding.bottom = x
        return self

    def set_footer_top_padding(self, x):
        self.__footer.style.padding.top = x
        return self

    # ============================================================
    # Items Section Settings
    # ============================================================

    def set_items_left_padding(self, x):
        self.__items_section.style.padding.left = x
        return self

    def set_items_right_padding(self, x):
        self.__items_section.style.padding.right = x
        return self

    def set_items_bottom_padding(self, x):
        self.__items_section.style.padding.bottom = x
        return self

    def set_items_top_padding(self, x):
        self.__items_section.style.padding.top = x
        return self

    def show_item_bottom_border(self, item_text, flag):
        # Allow a menu item to be passed in instead of text.
        if not isinstance(item_text, str) and hasattr(item_text, 'text'):
            item_text = item_text.text
        self.__items_section.show_item_bottom_border(item_text, flag)
        return self

    def show_item_top_border(self, item_text, flag):
        # Allow a menu item to be passed in instead of text.
        if not isinstance(item_text, str) and hasattr(item_text, 'text'):
            item_text = item_text.text
        self.__items_section.show_item_top_border(item_text, flag)
        return self

    # ============================================================
    # Prologue Section Settings
    # ============================================================

    def set_prologue_text_align(self, align='left'):
        self.__prologue.text_align = align
        return self

    def show_prologue_top_border(self, flag):
        self.__prologue.show_top_border = flag
        return self

    def show_prologue_bottom_border(self, flag):
        self.__prologue.show_bottom_border = flag
        return self

    # ============================================================
    # Epilogue Section Settings
    # ============================================================

    def set_epilogue_text_align(self, align='left'):
        self.__epilogue.text_align = align
        return self

    def show_epilogue_top_border(self, flag):
        self.__epilogue.show_top_border = flag
        return self

    def show_epilogue_bottom_border(self, flag):
        self.__epilogue.show_bottom_border = flag
        return self

    # ============================================================
    # Prompt Settings
    # ============================================================

    def set_prompt(self, prompt):
        self.__prompt.prompt = prompt
        return self

    # ============================================================
    # Menu generation
    # ============================================================

    def clear_data(self):
        """
        Clear menu data from previous menu generation.
        """
        self.__header.title = None
        self.__header.subtitle = None
        self.__prologue.text = None
        self.__epilogue.text = None
        self.__items_section.items = None

    def format(self, title=None, subtitle=None, prologue_text=None, epilogue_text=None, items=None):
        """
        Format the menu and return as a string.
        :return:  a string representation of the formatted menu.
        """
        self.clear_data()
        content = ''
        # Header Section
        if title is not None:
            self.__header.title = title
        if subtitle is not None:
            self.__header.subtitle = subtitle
        sections = [self.__header]
        # Prologue Section
        if prologue_text is not None:
            self.__prologue.text = prologue_text
            sections.append(self.__prologue)
        # Items Section
        if items is not None:
            self.__items_section.items = items
            sections.append(self.__items_section)
        # Epilogue Section
        if epilogue_text is not None:
            self.__epilogue.text = epilogue_text
            sections.append(self.__epilogue)
        sections.append(self.__footer)
        sections.append(self.__prompt)
        for sect in sections:
            content += "\n".join(sect.generate())
            # Don't add newline to prompt so input is on same line as prompt
            if not isinstance(sect, MenuPrompt):
                content += "\n"
        return content



from typing import Any, Optional

import pywintypes
import win32file
import win32pipe
from docker import DockerClient

from .....foundation.manager.terminal.core.ITerminalSession import ITerminalSession


class DockerNPipeSession(ITerminalSession):
    __slots__ = ['_exec_id', '_pipe_handle']

    def __init__(self, handler: Any, client: DockerClient, exec_id: str) -> None:
        super().__init__(handler, client)

        self._exec_id: str = exec_id

        self._pipe_handle: Optional[int] = handler._handle.handle
        win32pipe.SetNamedPipeHandleState(self._pipe_handle, 0x00000001, None, None)

    def read(self, n: int = 4096) -> bytes:
        """Read up to n bytes from the session output stream.

        Args:
            n (int): Maximum number of bytes to read.

        Returns:
            bytes: Data read from the session.
        """
        if self._closed or self._pipe_handle is None:
            return b""

        try:
            _, data = win32file.ReadFile(self._pipe_handle, n, None)
            return data
        except pywintypes.error as e:
            if getattr(e, "winerror", None) in (232, 233):
                return b""
            raise e

    def write(self, data: bytes) -> None:
        """Write bytes to the session input stream.

        Args:
            data (bytes): Data to send to the session.

        Returns:
            None
        """
        if self._closed or self._pipe_handle is None:
            return

        win32file.WriteFile(self._pipe_handle, data)

    def resize(self, cols: int, rows: int) -> None:
        """Resize the session terminal dimensions.

        Args:
            cols (int): Terminal width in columns.
            rows (int): Terminal height in rows.

        Returns:
            None
        """
        if self._closed:
            return

        self._client.api.exec_resize(self._exec_id, height=rows, width=cols)

    def close(self) -> None:
        """Close the session and release resources.

        Returns:
            None
        """
        if self._closed:
            return

        self._closed = True

        if self._pipe_handle is not None:
            try:
                win32file.CloseHandle(self._pipe_handle)
            except pywintypes.error:
                pass
            self._pipe_handle = None
